"""Data models for the Open Energy ID."""

import datetime as dt
from typing import Optional, overload, Union

try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

import pandas as pd
from pydantic import BaseModel


class TimeSeriesBase(BaseModel):
    """Pydantic base model for time series data."""

    index: list[dt.datetime]

    @classmethod
    def from_pandas(cls, data: Union[pd.Series, pd.DataFrame]) -> Self:
        """Create from a Pandas Object."""
        raise NotImplementedError

    def to_pandas(self, timezone: str = "UTC") -> Union[pd.Series, pd.DataFrame]:
        """Convert to a Pandas Object."""
        raise NotImplementedError

    @overload
    def to_json(self, path: None = None, **kwargs) -> str:
        """Dump to a JSON string."""

    @overload
    def to_json(self, path: str, **kwargs) -> None:
        """Dump to a JSON file."""

    def to_json(self, path: Optional[str] = None, **kwargs) -> Optional[str]:
        """Dump to a JSON string or file."""
        if path is None:
            return self.model_dump_json(**kwargs)
        encoding = kwargs.pop("encoding", "UTF-8")
        with open(path, "w", encoding=encoding) as file:
            file.write(self.model_dump_json(**kwargs))
        return None

    @overload
    @classmethod
    def from_json(cls, string: str, **kwargs) -> Self:
        """Load from a JSON string."""

    @overload
    @classmethod
    def from_json(cls, path: str, **kwargs) -> Self:
        """Load from a JSON file."""

    @classmethod
    def from_json(cls, string: Optional[str] = None, path: Optional[str] = None, **kwargs) -> Self:
        """Load from a JSON file or string."""
        if string:
            return cls.model_validate_json(string, **kwargs)
        if path:
            encoding = kwargs.pop("encoding", "UTF-8")
            with open(path, "r", encoding=encoding) as file:
                return cls.model_validate_json(file.read(), **kwargs)
        raise ValueError("Either string or path must be provided.")


class TimeSeries(TimeSeriesBase):
    """Time series data with a single column."""

    name: Union[str, None] = None
    data: list[float]

    @classmethod
    def from_pandas(cls, data: pd.Series) -> Self:
        """Create from a Pandas Series."""
        return cls.model_construct(name=data.name, data=data.tolist(), index=data.index.tolist())

    def to_pandas(self, timezone: str = "UTC") -> pd.Series:
        """Convert to a Pandas Series."""
        series = pd.Series(self.data, name=self.name, index=self.index)
        series.index = pd.to_datetime(series.index, utc=True)
        return series.tz_convert(timezone)


class TimeDataFrame(TimeSeriesBase):
    """Time series data with multiple columns."""

    columns: list[str]
    data: list[list[float]]

    @classmethod
    def from_pandas(cls, data: pd.DataFrame) -> Self:
        """Create from a Pandas DataFrame."""
        return cls.model_construct(
            columns=data.columns.tolist(), data=data.values.tolist(), index=data.index.tolist()
        )

    def to_pandas(self, timezone: str = "UTC") -> pd.DataFrame:
        """Convert to a Pandas DataFrame."""
        frame = pd.DataFrame(self.data, columns=self.columns, index=self.index)
        frame.index = pd.to_datetime(frame.index, utc=True)
        return frame.tz_convert(timezone)

    @classmethod
    def from_timeseries(cls, data: list[TimeSeries]) -> Self:
        """Create from a list of TimeSeries objects."""
        return cls.model_construct(
            columns=[series.name for series in data],
            data=[series.data for series in data],
            index=data[0].index,
        )

    def to_timeseries(self) -> list[TimeSeries]:
        """Convert to a list of TimeSeries objects."""
        return [
            TimeSeries(name=column, data=column_data, index=self.index)
            for column, column_data in zip(self.columns, self.data)
        ]



"""Models for dynamic tariff analysis."""

from typing import Literal
from pydantic import Field, conlist

from openenergyid.models import TimeDataFrame


RequiredColumns = Literal[
    "electricity_delivered",
    "electricity_exported",
    "price_electricity_delivered",
    "price_electricity_exported",
    "RLP",
    "SPP",
]

OutputColumns = Literal[
    "electricity_delivered_smr3",
    "electricity_exported_smr3",
    "price_electricity_delivered",
    "price_electricity_exported",
    "RLP",
    "SPP",
    "electricity_delivered_smr2",
    "electricity_exported_smr2",
    "cost_electricity_delivered_smr2",
    "cost_electricity_exported_smr2",
    "cost_electricity_delivered_smr3",
    "cost_electricity_exported_smr3",
    "rlp_weighted_price_delivered",
    "spp_weighted_price_exported",
    "heatmap_delivered",
    "heatmap_exported",
    "heatmap_total",
]


class DynamicTariffAnalysisInput(TimeDataFrame):
    """Input frame for dynamic tariff analysis."""

    columns: list[RequiredColumns] = Field(
        min_length=len(RequiredColumns.__args__),
        max_length=len(RequiredColumns.__args__),
        examples=[RequiredColumns.__args__],
    )
    data: list[
        conlist(
            item_type=float,
            min_length=len(RequiredColumns.__args__),
            max_length=len(RequiredColumns.__args__),
        )  # type: ignore
    ] = Field(examples=[[0.0, 0.0, 0.0, 0.0, 0.0, 0.0]])


class DynamicTariffAnalysisOutput(TimeDataFrame):
    """Output frame for dynamic tariff analysis."""

    columns: list[OutputColumns] = Field(
        min_length=1,
        max_length=len(OutputColumns.__args__),
        examples=[OutputColumns.__args__],
    )
    data: list[
        conlist(item_type=float, min_length=1, max_length=len(OutputColumns.__args__))  # type: ignore
    ] = Field(examples=[[0.0, 0.0, 0.0, 0.0, 0.0, 0.0]])



"""Constants for the energysharing module."""

GROSS_INJECTION = "Gross Injection"
NET_INJECTION = "Net Injection"
GROSS_OFFTAKE = "Gross Offtake"
NET_OFFTAKE = "Net Offtake"
KEY = "Key"
SHARED_ENERGY = "Shared Energy"



"""Data models for energy sharing."""

from enum import Enum
from typing import Annotated

from pydantic import BaseModel, Field
import pandas as pd

from openenergyid import TimeDataFrame
from .data_formatting import create_multi_index_input_frame
from .const import NET_INJECTION, NET_OFFTAKE, SHARED_ENERGY


class CalculationMethod(Enum):
    """Calculation method for energy sharing."""

    FIXED = "Fixed"
    RELATIVE = "Relative"
    OPTIMAL = "Optimal"


class EnergySharingInput(BaseModel):
    """Input data for energy sharing."""

    gross_injection: Annotated[TimeDataFrame, Field(alias="grossInjection")]
    gross_offtake: Annotated[TimeDataFrame, Field(alias="grossOfftake")]
    key: Annotated[TimeDataFrame, Field(alias="key")]
    timezone: str = Field(alias="timeZone", default="Europe/Brussels")

    def data_frame(self) -> pd.DataFrame:
        """Return the data as a combined DataFrame"""
        df = create_multi_index_input_frame(
            gross_injection=self.gross_injection.to_pandas(),
            gross_offtake=self.gross_offtake.to_pandas(),
            key=self.key.to_pandas(),
        )
        df = df.tz_convert(self.timezone)
        return df


class EnergySharingOutput(BaseModel):
    """Output data for energy sharing."""

    net_injection: TimeDataFrame = Field(alias="netInjection")
    net_offtake: TimeDataFrame = Field(alias="netOfftake")
    shared_energy: TimeDataFrame = Field(alias="sharedEnergy")

    @classmethod
    def from_calculation_result(cls, result: pd.DataFrame) -> "EnergySharingOutput":
        """Create an output model from a calculation result."""
        return cls.model_construct(
            net_injection=TimeDataFrame.from_pandas(result[NET_INJECTION]),
            net_offtake=TimeDataFrame.from_pandas(result[NET_OFFTAKE]),
            shared_energy=TimeDataFrame.from_pandas(result[SHARED_ENERGY]),
        )



"""Energy Sharing package."""

from .main import calculate
from .models import CalculationMethod, EnergySharingInput, EnergySharingOutput

__all__ = ["calculate", "CalculationMethod", "EnergySharingInput", "EnergySharingOutput"]



"""Main module of the DynTar package."""

from typing import Optional
import pandas as pd

from openenergyid.const import (
    ELECTRICITY_DELIVERED,
    ELECTRICITY_EXPORTED,
    PRICE_ELECTRICITY_DELIVERED,
    PRICE_ELECTRICITY_EXPORTED,
    RLP,
    SPP,
)

from .const import (
    ELECTRICITY_DELIVERED_SMR3,
    ELECTRICITY_EXPORTED_SMR3,
    ELECTRICITY_DELIVERED_SMR2,
    ELECTRICITY_EXPORTED_SMR2,
    COST_ELECTRICITY_DELIVERED_SMR2,
    COST_ELECTRICITY_EXPORTED_SMR2,
    COST_ELECTRICITY_DELIVERED_SMR3,
    COST_ELECTRICITY_EXPORTED_SMR3,
    RLP_WEIGHTED_PRICE_DELIVERED,
    SPP_WEIGHTED_PRICE_EXPORTED,
    HEATMAP_DELIVERED,
    HEATMAP_EXPORTED,
    HEATMAP_TOTAL,
)


def weigh_by_monthly_profile(series: pd.Series, profile: pd.Series) -> pd.Series:
    """Weigh a time series by a monthly profile."""
    df = pd.DataFrame({"series": series, "profile": profile})
    results = []
    for _, frame in df.groupby(pd.Grouper(freq="MS")):
        frame = frame.copy()
        frame["weighted"] = frame["series"].sum() * (frame["profile"] / frame["profile"].sum())
        results.append(frame)
    return pd.concat(results)["weighted"]


def extend_dataframe_with_smr2(df: pd.DataFrame, inplace: bool = False) -> Optional[pd.DataFrame]:
    """Extend a DataFrame with the SMR2 columns."""
    if not inplace:
        result_df = df.copy()
    else:
        result_df = df

    result_df[ELECTRICITY_DELIVERED_SMR2] = weigh_by_monthly_profile(
        df[ELECTRICITY_DELIVERED], df[RLP]
    )
    result_df[ELECTRICITY_EXPORTED_SMR2] = weigh_by_monthly_profile(
        df[ELECTRICITY_EXPORTED], df[SPP]
    )

    result_df.rename(
        columns={
            ELECTRICITY_DELIVERED: ELECTRICITY_DELIVERED_SMR3,
            ELECTRICITY_EXPORTED: ELECTRICITY_EXPORTED_SMR3,
        },
        inplace=True,
    )

    if not inplace:
        return result_df
    return None


def extend_dataframe_with_costs(df: pd.DataFrame, inplace: bool = False) -> Optional[pd.DataFrame]:
    """Extend a DataFrame with the cost columns."""
    if not inplace:
        result_df = df.copy()
    else:
        result_df = df

    result_df[COST_ELECTRICITY_DELIVERED_SMR2] = (
        df[ELECTRICITY_DELIVERED_SMR2] * df[PRICE_ELECTRICITY_DELIVERED]
    )
    result_df[COST_ELECTRICITY_EXPORTED_SMR2] = (
        df[ELECTRICITY_EXPORTED_SMR2] * df[PRICE_ELECTRICITY_EXPORTED]
    )

    result_df[COST_ELECTRICITY_DELIVERED_SMR3] = (
        df[ELECTRICITY_DELIVERED_SMR3] * df[PRICE_ELECTRICITY_DELIVERED]
    )
    result_df[COST_ELECTRICITY_EXPORTED_SMR3] = (
        df[ELECTRICITY_EXPORTED_SMR3] * df[PRICE_ELECTRICITY_EXPORTED]
    )

    if not inplace:
        return result_df
    return None


def extend_dataframe_with_weighted_prices(
    df: pd.DataFrame, inplace: bool = False
) -> Optional[pd.DataFrame]:
    """Extend a DataFrame with the weighted price columns."""
    if not inplace:
        df = df.copy()

    rlp_weighted_price_delivered = (df[PRICE_ELECTRICITY_DELIVERED] * df[RLP]).resample(
        "MS"
    ).sum() / df[RLP].resample("MS").sum()
    df[RLP_WEIGHTED_PRICE_DELIVERED] = rlp_weighted_price_delivered.reindex_like(
        df[RLP], method="ffill"
    )
    spp_weighted_price_exported = (df[PRICE_ELECTRICITY_EXPORTED] * df[SPP]).resample(
        "MS"
    ).sum() / df[SPP].resample("MS").sum()
    df[SPP_WEIGHTED_PRICE_EXPORTED] = spp_weighted_price_exported.reindex_like(
        df[SPP], method="ffill"
    )

    if not inplace:
        return df
    return None


def extend_dataframe_with_heatmap(
    df: pd.DataFrame, inplace: bool = False
) -> Optional[pd.DataFrame]:
    """Extend a DataFrame with the heatmap columns."""
    if not inplace:
        df = df.copy()

    heatmap_score_delivered = (
        (df[ELECTRICITY_DELIVERED_SMR2] - df[ELECTRICITY_DELIVERED_SMR3])
        / df[ELECTRICITY_DELIVERED_SMR2]
        * (df[RLP_WEIGHTED_PRICE_DELIVERED] - df[PRICE_ELECTRICITY_DELIVERED])
        / df[RLP_WEIGHTED_PRICE_DELIVERED]
    )
    heatmap_score_exported = (
        (df[ELECTRICITY_EXPORTED_SMR2] - df[ELECTRICITY_EXPORTED_SMR3])
        / df[ELECTRICITY_EXPORTED_SMR2]
        * (df[SPP_WEIGHTED_PRICE_EXPORTED] - df[PRICE_ELECTRICITY_EXPORTED])
        / df[SPP_WEIGHTED_PRICE_EXPORTED]
    )
    heatmap_score_delivered.fillna(0, inplace=True)
    heatmap_score_exported.fillna(0, inplace=True)

    # Invert scores so that positive values indicate a positive impact
    heatmap_score_delivered = -heatmap_score_delivered
    heatmap_score_exported = -heatmap_score_exported
    heatmap_score_combined = heatmap_score_delivered - heatmap_score_exported

    df[HEATMAP_DELIVERED] = heatmap_score_delivered
    df[HEATMAP_EXPORTED] = heatmap_score_exported
    df[HEATMAP_TOTAL] = heatmap_score_combined

    if not inplace:
        return df
    return None


def calculate_dyntar_columns(df: pd.DataFrame, inplace: bool = False) -> Optional[pd.DataFrame]:
    """Calculate all columns required for the dynamic tariff analysis."""
    if not inplace:
        df = df.copy()

    extend_dataframe_with_smr2(df, inplace=True)
    extend_dataframe_with_costs(df, inplace=True)
    extend_dataframe_with_weighted_prices(df, inplace=True)
    extend_dataframe_with_heatmap(df, inplace=True)

    if not inplace:
        return df
    return None
